# Copyright (c) 2023 Alexander Hauck. Distributed under the terms of the MIT License.

"""
findTDE comprises a set of scripts to facilitate easy, high-throughput calculations of threshold displacement energies for materials in VASP/LAMMPS using ab initio/classical molecular dynamics.
"""

__author__ = "Alexander Hauck"
__copyright__ = "Copyright (C) 2023 Alexander Hauck"
__license__ = "MIT License"
__version__ = "1.1.0"
